import base64
import quopri
from copy import copy
from datetime import datetime
from email.header import decode_header

from bs4 import BeautifulSoup


class DecodeException(Exception):
    pass


def get_decode_content(encode_str):
    """
    获取邮件解码后的内容
    :param encode_str:
    :return:
    """
    decode_str, encoding = decode_header(encode_str)[0]
    # decode 失败处理
    if decode_str == b'"' and '=' in encode_str:
        encode_str = encode_str.replace('"=', '=').replace('="', '=')
        decode_str, encoding = decode_header(encode_str)[0]

    if encoding == 'unknown-8bit' or encoding == 'gb2312':
        # 解码失败默认用 gbk
        encoding = 'gbk'

    if encoding is not None:
        decode_str = decode_str.decode(encoding)

    if '�' in decode_str:
        raise DecodeException('Decode_content_error.')
    return decode_str


def get_subject(msg):
    """
    获取邮件主题
    :param msg: Message 对象
    :return:
    """
    return get_decode_content(msg.get('Subject'))


def get_mail_date(msg):
    """
    获取邮件日期
    :param msg: Message 对象
    :return:
    """
    # https://www.simplifiedpython.net/python-convert-string-to-datetime/
    # 'Wed, 25 Sep 2019 20:59:23 +0800' ==> '25 Sep 2019 20:59:23'
    date_str = msg.get('Date').split(',')[-1].split('+')[0].strip()
    try:
        # 国内邮件时间解析
        return datetime.strptime(date_str, '%d %b %Y %H:%M:%S')
    except ValueError:
        # 英文系统邮件时间解析
        # 'Tue Jan 21 18:00:09 2020'
        return datetime.strptime(date_str, '%a %b %d %H:%M:%S %Y')


def get_sender(msg):
    """
    获取邮件发件人名称，不是发件人邮件。
    '"eservice@nbcb.cn" <eservice@nbcb.cn>' => eservice@nbcb.cn
    '<yywbfa@cmschina.com.cn>' => yywbfa
    '=?GB2312?B?uqPNqNakyK8=?= <ppos@htsec.com>' => 海通证券
    :param msg: Message 对象
    :return:
    """
    raw_sender = msg.get('From')
    if '?' in raw_sender:
        sender = get_decode_content(raw_sender)
    else:
        if '"' in raw_sender:
            start = raw_sender.find('"') + 1
            end = raw_sender.find('"', start)
            sender = raw_sender[start:end]
        else:
            sender = raw_sender.split('@')[0]
            sender = sender.strip('<>')
    return sender


def has_attachment(msg):
    """
    check whether has attachments
    :param msg:
    :return:
    """
    attachment_num = msg.get('X-ATTACHMENT-NUM')
    # plain text 也算 attachment
    if attachment_num < 2:
        return False
    else:
        return True


def get_attachment_name(part):
    """
    get chinese attachment name if part.get_filename() fails.
    :param part: Message object
    :return:
    """
    file_name = part.get_filename()
    try:
        file_name = get_decode_content(file_name)
        return file_name
    except (AttributeError, DecodeException) as e:
        # Use shallow copy avoid string huge payload
        part_shallow = copy(part)
        part_shallow._payload = None
        part_str = part_shallow.as_string()
        part_str_elements = part_str.split('unknown-8bit')
        for item in part_str_elements:
            if '?b?' in item:
                file_name_b64 = item.split('?=')[0].strip().replace('?b?', '')
                for encoding in ('gbk', 'gb2312', 'utf-8'):
                    try:
                        file_name = base64.b64decode(file_name_b64).decode(encoding)
                    except UnicodeDecodeError:
                        continue
                file_name = file_name.split('=')[-1].replace('"', '')
                return file_name


def get_attachment(msg, file_filter=('xls', 'xlsx', 'pdf', 'doc', 'docx')):
    """
    获取邮件里的附件
    :param msg:
    :param file_filter:
    :return:
    """
    attachments = []
    for part in msg.walk():
        if part.get_content_maintype() == 'multipart': continue
        if part.get_content_maintype() == 'text': continue
        if part.get('Content-Disposition') is None: continue
        file_name = get_attachment_name(part)
        file_extension = file_name.split('.')[-1]
        file = part.get_payload(decode=True)
        if file_extension not in file_filter:
            continue
        attachments.append({'name': file_name, 'file': file})
    return attachments


def get_body_content(msg):
    """
    获取邮件 body 信息
    :param msg:
    :return:
    """
    content = None
    for part in msg.walk():
        if part.get_content_type() in ("text/html", "text/plain"):
            content_encode = part.get_payload()
            content_transfer_encoding = part.get('Content-Transfer-Encoding')
            if part.get('Content-Type'):
                charset = part.get('Content-Type').split('=')[1].lower()
                if content_transfer_encoding == 'quoted-printable':
                    content = quopri.decodestring(content_encode).decode(charset)
                elif content_transfer_encoding == 'base64':
                    try:
                        content = base64.b64decode(content_encode).decode(charset)
                    except UnicodeDecodeError as e:
                        # 可能是这种 '"gb2312"'，所以用 in
                        if 'gb2312' in charset:
                            content = base64.b64decode(content_encode).decode('gbk')
                        else:
                            raise e
                elif content_transfer_encoding == '7bit' or content_transfer_encoding == '8bit':
                    content = content_encode
                else:
                    raise Exception('Content 无法处理。')

                if part.get_content_type() == 'text/html':
                    content = {'text': BeautifulSoup(content, features='lxml').text, 'html': content}
                else:
                    content = {'text': content}

    return content



import email
import ssl

import requests
from imapclient import IMAPClient, SEEN
from imapclient.exceptions import IMAPClientError
from tqdm import tqdm

from pymail_aliyun.mail_utils import get_subject, get_mail_date, get_body_content, get_sender, get_attachment

USR_AGENT = (
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) '
    'AppleWebKit/537.36 (KHTML, like Gecko) '
    'Chrome/78.0.3904.108 Safari/537.36'
)

ssl_context = ssl.create_default_context()
# don't check if certificate hostname doesn't match target hostname
ssl_context.check_hostname = False
# don't check if the certificate is trusted by a certificate authority
ssl_context.verify_mode = ssl.CERT_NONE


class MailAliyun(object):
    """
    单封邮件对像
    """

    def __init__(self, uid, subject, mail_date, body, attachments, sender):
        self.subject = subject
        self.mail_date = mail_date
        self.body = body
        self.attachments = attachments
        self.uid = uid
        self.sender = sender


class MailAliyunClient(IMAPClient):
    """
    Extend IMAPClient with aliyun mail
    """

    def __init__(self, user_name, password, host='imap.aliyun.com', port=None, use_uid=True, ssl=True, stream=False,
                 ssl_context=ssl_context, timeout=None):
        super(MailAliyunClient, self).__init__(host, port=port, use_uid=use_uid, ssl=ssl, stream=stream,
                                               ssl_context=ssl_context, timeout=timeout)
        self._user_name = user_name
        self._password = password
        self.req_session = requests.Session()

    def __enter__(self):
        self.login(self._user_name, self._password)
        return self

    def mark_unread(self, uid):
        """
        标记邮件为未读
        :param uid: mail uid
        :return:
        """
        self.remove_flags(uid, [SEEN])

    def mark_read(self, uid):
        """
        标记邮件为已读
        :param uid:
        :return:
        """
        self.add_flags(uid, [SEEN])

    def create_aliyun_folder(self, folder_name):
        """
        创建目录，支持 nested 目录。
        :return: None or {folder_name: folder_id}
        """
        paths_element = folder_name.split('/')
        possible_path = ['/'.join(paths_element[:i]) for i in reversed(range(1, len(paths_element) + 1))]
        final_pointer = 0
        total_iter_counts = 0
        while True:
            if total_iter_counts > 10:
                raise Exception('Too many nested folder')
            try:
                msg = self.create_folder(possible_path[final_pointer])
                if msg.decode('utf-8') == 'CREATE completed':
                    if final_pointer == 0:
                        break
                    final_pointer = final_pointer - 1
            except IMAPClientError:
                # 失败后，尝试创建父目录。
                final_pointer = final_pointer + 1
            total_iter_counts = total_iter_counts + 1

    def move_mail(self, uid, dest_path):
        """
        :param uid:
        :param dest_path: 准备移动到的 '/parent_folder/sub_folder/sub_folder'
        :return:
        """
        dest_path = dest_path.strip('/')
        try:
            self.copy(uid, dest_path)
        except IMAPClientError:
            # 可能是没有目标文件夹，或者此邮件就在这个目标文件夹里。
            if self.folder_exists(dest_path):
                return
            else:
                self.create_aliyun_folder(dest_path)
                self.copy(uid, dest_path)

    def handle_mails(self, func, folder_name='Inbox', criteria='UNSEEN', mails_count=50,
                     attch_filter=('xls', 'xlsx', 'pdf', 'doc', 'docx'), **kwargs):
        """
        处理未读邮件
        :param func: 处理邮件的方法，input 为 Mail263 对象，output 为 'success' or 'fail'
        :param folder_name: 邮件所在文件夹
        :param criteria: 搜索条件
        :param mails_count: 设置一次处理多少邮件
        :param attch_filter: 附件过滤器，只保留哪些附件
        :param kwargs: nas_client nas 客户端等 func 需要用到的客户端
        :return:
        """
        self.select_folder(folder_name)
        messages = self.search(criteria)
        if messages:
            if mails_count != 'ALL':
                messages = messages[:mails_count]
            for uid, message_data in tqdm(self.fetch(messages, 'RFC822').items()):
                email_message = email.message_from_bytes(message_data[b'RFC822'])
                try:
                    subject = get_subject(email_message)
                except Exception:
                    self.move_mail(uid, 'Decode_failed/subject')
                    break

                try:
                    mail_date = get_mail_date(email_message)
                except Exception:
                    self.move_mail(uid, 'Decode_failed/mail_date')
                    break

                try:
                    body = get_body_content(email_message)
                except Exception:
                    self.move_mail(uid, 'Decode_failed/body')
                    break

                try:
                    attachments = get_attachment(email_message, file_filter=attch_filter) or None
                except Exception:
                    self.move_mail(uid, 'Decode_failed/attachments')
                    break

                try:
                    sender = get_sender(email_message)
                except Exception:
                    self.move_mail(uid, 'Decode_failed/sender')
                    break

                mail = MailAliyun(uid, subject, mail_date, body, attachments, sender)
                func(self, mail, **kwargs)
        self.close_folder()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pymail_aliyun']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.8.1,<5.0.0',
 'imapclient>=2.1.0,<3.0.0',
 'optionaldict>=0.1.1,<0.2.0',
 'requests>=2.22.0,<3.0.0',
 'selenium>=3.141.0,<4.0.0',
 'simplejson>=3.17.0,<4.0.0',
 'tqdm>=4.41.1,<5.0.0']

setup_kwargs = {
    'name': 'pymail-aliyun',
    'version': '0.1.14',
    'description': 'mail-aliyun wrapper',
    'long_description': None,
    'author': 'zbjdonald',
    'author_email': 'service@yingchengtz.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
